Object subclass: #Logic
	instanceVariableNames: 'answer varList unboundVars tempResult returnBlock'
	classVariableNames: 'BinaryOp Database'
	poolDictionaries: ''
	category: 'Prolog-Base'!

!Logic methodsFor: 'query' stamp: 'mt 6/28/1999 12:00'!
buildQuery: aList
        "Private"
    | pred block |
    pred := aList head.
    self addVarFrom: pred tail.
    block :=
        (aList tail isEmpty
            ifTrue: [[self inferAnswer]]
            ifFalse: [self buildQuery: aList tail]).
    ^[self perform: pred head with:
        (Association
            key: pred tail
            value: block)]! !

!Logic methodsFor: 'query' stamp: 'spfa 6/8/2008 21:17'!
doGoal: aRelation continue: aBlock
	"Try to satisfy aRelation as a goal."
	| blk mark |
	mark := unboundVars size.
	((aRelation isKindOf: List) and: [
	aRelation head isSymbol])
		ifFalse: [self error: 'not a goal'].
	blk := self
		perform: (Logic selectorForPredicate: aRelation head)
		with: (Association
			key: aRelation tail
			value: aBlock).
	[blk class == BlockContext]
		whileTrue: [blk := blk value].
	self unbindVar: mark! !

!Logic methodsFor: 'query' stamp: 'spfa 6/9/2008 08:57'!
doOneGoal: aRelation
	"Try to satisfy aRelation as a goal.
	 Answer true if success, else false."
	| blk mark |
	mark := unboundVars size.
	((aRelation isKindOf: List) and: [
	aRelation head isSymbol])
		ifFalse: [self error: 'not a goal'].
	blk := self
		perform: (Logic selectorForPredicate: aRelation head)
		with: (Association
			key: aRelation tail
			value: [^true]).
	[blk class == BlockContext]
		whileTrue: [blk := blk value].
	self unbindVar: mark.
	^false! !

!Logic methodsFor: 'query' stamp: 'mt 6/28/1999 12:00'!
match: x with: y
        "Private -- Answer true if x matches y, else
         answer false."
    | xLast yLast |
 "Transcript nextPutAll: x printString;nextPut: $ ;
 nextPutAll: y printString;cr."
    (x class == LogicDontCare or: [y class == LogicDontCare])
         ifTrue: [^true].
    (x class == LogicRef)
        ifTrue: [
            xLast := x lastNode.
            xLast value class == LogicUnbound
                ifTrue: [
                    unboundVars add: xLast.
                    xLast value: y.
                    ^true]
                ifFalse: [xLast := xLast value]]
        ifFalse: [xLast := x].
    (y class == LogicRef)
        ifTrue: [
            yLast := y lastNode.
            yLast value class == LogicUnbound
                ifTrue: [
                    unboundVars add: yLast.
                    yLast value: x.
                    ^true]
                ifFalse: [
                    yLast := yLast value]]
        ifFalse: [yLast := y].
    (xLast class == List) ifTrue: [
        (yLast class == List)
            ifTrue: [
                (self match: xLast head with: yLast head)
                    ifTrue: [^self
                        match: xLast tail
                        with: yLast tail]]
            ifFalse: [^false]].
    (xLast class == Relation) ifTrue: [
        (yLast class == Relation)
            ifTrue: [
                (self match: xLast head with: yLast head)
                    ifTrue: [^self
                        match: xLast tail
                        with: yLast tail]]
            ifFalse: [^false]].
    ^xLast = yLast! !

!Logic methodsFor: 'query' stamp: 'mt 6/28/1999 12:00'!
retrieve: assoc from: dataList
        "Retrieve a fact in assoc from
         Database entry dataList."
    | aList |
    dataList isEmpty ifTrue: [^self].
    aList := dataList.
    [aList isLastNode] whileFalse: [
        self unify: assoc key
            with: aList head
            then: assoc value.
        aList := aList tail].
    self unify: assoc key
        with: aList head
        then: [^assoc value]! !

!Logic methodsFor: 'query' stamp: 'bolot 7/28/1999 23:09'!
unify: x with: y then: aBlock
        "Private -- If x unifies with y, then execute
         aBlock. Answer self."
    | mark |
"Transcript show: x printString, ' ', y printString; cr."
    mark := unboundVars size.
    (self match: x with: y)
        ifTrue:
            [tempResult := aBlock.
             [(tempResult := tempResult value).
              tempResult class == BlockContext]
                whileTrue: []].
    self unbindVar: mark! !


!Logic methodsFor: 'accessing' stamp: 'spfa 6/8/2008 20:49'!
addVarFrom: aList
"
  SH: 2004/11/21 BugFix
  SH: There was a problem in using include: here.....fixed 041026
"
    aList isEmpty ifTrue: [^self].

    (aList head isKindOf: List)
        ifTrue: [self addVarFrom: aList head]
        ifFalse: [
            ((aList head class == LogicRef and:
            [aList head value class == LogicUnbound]) and:
            [(varList includes: aList head) not])
                ifTrue: [
				varList add: aList head.
				]
			].
    self addVarFrom: aList tail! !

!Logic methodsFor: 'accessing' stamp: 'programmatic 6/30/1999 19:05'!
allValue: aLogicRef
        "Copy all the values to an instance
         the same class as the content of aLogicRef.
         If aLogicRef's content is a List, walk
         through all elements."
    | anAnswer aList |
    (aLogicRef isKindOf: LogicRef)
        ifTrue: [anAnswer := aLogicRef value]
        ifFalse: [anAnswer := aLogicRef].
    (anAnswer isKindOf: List)
        ifTrue: [
            anAnswer isEmpty ifTrue: [^anAnswer].
            aList := anAnswer.
            anAnswer := anAnswer class with:
                (self allValue: aList head).
            anAnswer tail: (self allValue: aList tail)].
    ^anAnswer! !

!Logic methodsFor: 'accessing' stamp: 'spfa 6/8/2008 19:31'!
doesNotUnderstand: aMessage
    | temp |

    (aMessage selector isUnary
		and: [self class canUnderstand: (temp _ ('prim', aMessage selector, ':') asSymbol)])
	ifTrue: [^ self perform: temp with: aMessage arguments first].

    temp := aMessage selector.
    temp := (temp copyFrom: 1 to: temp size - 1)
        asSymbol.
    temp := Database at: temp ifAbsent: [
        ^self error: 'Message not understood: ', aMessage selector
        ].
    ^self retrieve: (aMessage arguments at: 1)
        from: temp! !

!Logic methodsFor: 'accessing' stamp: 'bolot 7/7/1999 19:52'!
structureHead: aSymbol arguments: anInteger
        "Answer a structure whose functor is
         aSymbol and arguments are anInteger
         number of DontCares."
    | aRelation |
    aRelation := Relation with: aSymbol.
    anInteger timesRepeat: [
        aRelation add: LogicDontCare new].
    ^aRelation! !

!Logic methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
value: aRef
        "Private -- Answer the value of the
         logic ref, aRef."
    aRef class == LogicRef
        ifTrue: [^aRef value]
        ifFalse: [^aRef]! !


!Logic methodsFor: 'private' stamp: 'programmatic 6/30/1999 19:04'!
first: assoc
        "Private -- Answer the first argument."
    | anAnswer |
    anAnswer := assoc key head.
    ^self value: anAnswer! !

!Logic methodsFor: 'private' stamp: 'bolot 8/1/1999 12:42'!
inferAnswer
        "Private -- Add an Array of values to
         the answer collection."
    | anArray i |
    anArray := Array new: varList size.
    i := 1.
    varList do: [:x |
        anArray at: i put: (self allValue: x).
        i := i + 1].
    answer add: anArray.! !

!Logic methodsFor: 'private' stamp: 'spfa 6/8/2008 20:50'!
inferSmalltalk: aList var: bList

"SH: 2004/11/21 BugFix"

        "Private -- Handle query from Smalltalk. aList
         contains a list of predicates and
         bList is a list of unbound variables."

    | blk |

    varList := List new.
    answer := OrderedCollection new.
    returnBlock := [
        self unbindVar: 0.
        answer size = 0
            ifTrue: [^nil]
            ifFalse: [^answer asArray]].
    blk := self buildQuery: aList.
    [(blk := blk value) class == BlockContext]
        whileTrue: [].
    ^returnBlock value! !

!Logic methodsFor: 'private' stamp: 'programmatic 6/30/1999 19:05'!
second: assoc
        "Private -- Answer the second argument."
    | anAnswer |
    anAnswer := assoc key tail head.
    ^self value: anAnswer! !

!Logic methodsFor: 'private' stamp: 'mt 6/28/1999 12:00'!
unbindVar: mark
        "Private -- Unbind previously bound variables
         down to mark."
    | index |
    [(index := unboundVars size) > mark]
        whileTrue: [
            (unboundVars at: index)
                content: LogicUnbound new.
            unboundVars removeLast]! !


!Logic methodsFor: 'initialization' stamp: 'mt 6/28/1999 12:00'!
initialize
        "Private -- Initialize instance variables."
    unboundVars := OrderedCollection new: 2000! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Logic class
	instanceVariableNames: ''!

!Logic class methodsFor: 'initialize-release' stamp: 'mt 6/28/1999 12:00'!
initialize
        "Initialize Database."
    Database := Dictionary new! !


!Logic class methodsFor: 'instance creation' stamp: 'mt 6/28/1999 12:00'!
new
    ^super new initialize! !


!Logic class methodsFor: 'accessing' stamp: 'spfa 6/8/2008 20:39'!
compilerClass
	^LCompiler! !

!Logic class methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
database
    ^Database! !


!Logic class methodsFor: 'as yet unclassified' stamp: 'spfa 6/8/2008 21:25'!
prologKeyword

	^ 'logic'
! !

!Logic class methodsFor: 'as yet unclassified' stamp: 'spfa 6/8/2008 21:20'!
selectorForPredicate: aString

	^ (self prologKeyword, (aString first: 1) asUppercase, aString allButFirst) asSymbol! !


Object subclass: #LogicDontCare
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Base'!

!LogicDontCare methodsFor: 'parser support' stamp: 'bolot 7/19/1999 17:55'!
size
	" symbol used for LogicDontCare is '_' (underscore) "
	^self printString size! !


!LogicDontCare methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
printOn: aStream
    aStream nextPutAll: '_';
        nextPut: $ ! !

!LogicDontCare methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
= anObject
    ^true! !


Object subclass: #LogicRef
	instanceVariableNames: 'content'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Base'!

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
at: anObject
    ^self value at: anObject! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
content
    ^content! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
content: anObject
    content := anObject! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
doesNotUnderstand: aMessage
    ^self value
        perform: aMessage selector
        withArguments: aMessage arguments! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
lastNode
    | aNode |
    aNode := self.
    [aNode content class == LogicRef]
        whileTrue: [aNode := aNode content].
    ^aNode! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
printOn: aStream
    self value printOn: aStream! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
value
    | aNode |
    aNode := content.
    [aNode class == LogicRef]
        whileTrue: [aNode := aNode content].
    ^aNode! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
value: anObject
    | node |
    node := self lastNode.
    anObject class == LogicRef
        ifTrue: [
            node == anObject lastNode
                ifFalse: [node content: anObject]]
        ifFalse: [
            node content: anObject]! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
= anObject
    ^self value = anObject! !

!LogicRef methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
== anObject
    ^self value == anObject! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

LogicRef class
	instanceVariableNames: ''!

!LogicRef class methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
new
    ^super new value: LogicUnbound new! !


Object subclass: #LogicUnbound
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog-Base'!

!LogicUnbound methodsFor: 'accessing' stamp: 'mt 6/28/1999 12:00'!
printOn: aStream
    'Unbound' printOn: aStream! !

Logic initialize!