Logic subclass: #Prolog
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog'!
!Prolog commentStamp: 'bolot 12/13/2002 15:12' prior: 0!
This class and other related classes in this category were originally written by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.  We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering.!


!Prolog methodsFor: 'complex goals' stamp: 'spfa 6/8/2008 21:28'!
primlogicAnd: assoc
    "Try to satisfy a list of goals."
    | aList |
    aList := assoc key.
    [aList isEmpty] whileFalse: [
        (self doOneGoal: aList head)
            ifTrue: [aList := aList tail]
            ifFalse: [^self]].
    ^assoc value! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 6/8/2008 21:35'!
primlogicCall: assoc
        "Execute the argument as a predicate."
    | blk |
    assoc key size = 1 ifFalse: [^self].
    blk := self first: assoc.
    ((blk isKindOf: List) and: [
    blk head isSymbol])
        ifFalse: [self error: 'not a goal'].
    ^[self perform: (blk head, ':') asSymbol
            with: (Association key: blk tail
                value: assoc value)]! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 6/8/2008 22:19'!
primlogicNot: assoc
        "Succeed if the evaluation of the argument
         as a goal fails."
    | blk |
    assoc key size = 1 ifFalse: [^self].
    blk := self first: assoc.
    (blk isKindOf: List)
        ifFalse: [self error: 'argument must be goal'].
    blk :=
        self perform: (Logic selectorForPredicate: blk head) 
            with:
                (Association key: blk tail
                    value: [^self]).
    [blk class == BlockContext] "value until real success"
        whileTrue: [blk := blk value].
    ^assoc value! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 6/8/2008 21:29'!
primlogicOr: assoc
    "Disjunction predicate."
    assoc key size = 2 ifFalse: [^self].
    self doGoal: (self first: assoc) continue: assoc value.
    self doGoal: (self second: assoc) continue: assoc value! !

!Prolog methodsFor: 'complex goals' stamp: 'spfa 6/8/2008 21:29'!
primlogicRepeat: assoc
        "Succeed always, even
         during backtacking."
    | blk mark |
    assoc key size = 0 ifFalse: [^self].
    mark := unboundVars size.
    [true] whileTrue: [
        blk := assoc value.
        [(blk := blk value) class == BlockContext]
            whileTrue: [].
        self unbindVar: mark]! !


!Prolog methodsFor: 'database' stamp: 'spfa 6/8/2008 21:29'!
primlogicAsserta: assoc
        "Add the fact in assoc to the
         beginning of the database."
    | arg aList |
    assoc key size = 1 ifFalse: [^self].
    arg := self allValue: (self first: assoc).
    aList := Database at: arg head ifAbsent: [^self].
    Database at: arg head put:
        (List head: arg tail tail: aList).
    ^assoc value! !

!Prolog methodsFor: 'database' stamp: 'spfa 6/8/2008 21:29'!
primlogicAssertz: assoc
        "Add the fact in assoc to the
         end of the database."
    | arg aList |
    assoc key size = 1 ifFalse: [^self].
    arg := self allValue: (self first: assoc).
    aList := Database at: arg head ifAbsent: [^self].
    aList add: arg tail.
    ^assoc value! !

!Prolog methodsFor: 'database' stamp: 'spfa 6/8/2008 21:29'!
primlogicPurgeDatabase: assoc
        "Purge a database identified by the
         structure in the argument."
    | arg |
    assoc key size = 1 ifFalse: [^self].
    arg := self first: assoc.
    ((arg isKindOf: Relation) and:
    [arg head isSymbol])
        ifFalse: [self error: 'argument is not a fact'].
    Database at: arg head ifAbsent: [^self].
    Database removeKey: arg head.
    ^assoc value! !

!Prolog methodsFor: 'database' stamp: 'spfa 6/8/2008 21:29'!
primlogicRetract: assoc
        "Delete the first fact unified with the
         one in assoc from the database. Delete
         the next one during backtrack."
    | arg aList bList |
    assoc key size = 1 ifFalse: [^self].
    arg := self first: assoc.
    [aList := Database at: arg head ifAbsent: [^self].
     aList isEmpty ifTrue: [^self].
     self unify: arg tail
        with: aList head
        then: [Database at: arg head put: aList tail.
               assoc value value].
     aList == (Database at: arg head)]
        whileFalse: [].
    [(bList := aList tail) isEmpty] whileFalse: [
        self unify: arg tail
            with: aList tail head
            then: [aList tail: aList tail tail.
                   bList := aList.
                   assoc value value].
        aList := bList.
        aList isEmpty ifTrue: [^self]]! !


!Prolog methodsFor: 'classifying terms' stamp: 'spfa 6/8/2008 21:29'!
primlogicAtomic: assoc
        "Succeed if the argument is a Symbol
         or an integer."
    | arg |
    assoc key size = 1 ifFalse: [^self].
    ((arg := self first: assoc) isSymbol
    or: [arg isKindOf: Integer])
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 6/8/2008 21:30'!
primlogicAtom: assoc
        "Succeed if the argument is a Symbol."
    assoc key size = 1 ifFalse: [^self].
    (self first: assoc) isSymbol
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 6/8/2008 21:30'!
primlogicInteger: assoc
        "Succeed if the argument is an Integer."
    assoc key size = 1 ifFalse: [^self].
    ((self first: assoc) isKindOf: Integer)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 6/8/2008 21:30'!
primlogicNonvar: assoc
        "Succeed if the argument is not an
         unbound variable."
    assoc key size = 1 ifFalse: [^self].
    (self first: assoc) class == LogicUnbound
        ifFalse: [^assoc value]! !

!Prolog methodsFor: 'classifying terms' stamp: 'spfa 6/8/2008 21:30'!
primlogicVar: assoc
        "Succeed if the argument is an
         unbound variable."
    assoc key size = 1 ifFalse: [^self].
    (self first: assoc) class == LogicUnbound
        ifTrue: [^assoc value]! !


!Prolog methodsFor: 'comm with smalltalk' stamp: 'spfa 6/8/2008 23:10'!
primlogicConsult: assoc
        "Ask the second argument as the receiver
         (a Smalltalk expression) to solve the goals
         contained in the first argument, e.g.
            consult(father(x, 'John'), family new)."
    | blk |
    assoc key size = 2 ifFalse: [^self].
    blk := self first: assoc.
    blk class == Relation
        ifFalse: [self error: 'first argument must be predicate'].
    ^[(self second: assoc) value
        perform: (Logic selectorForPredicate: blk head) 
        with: (Association key: blk tail
               value: assoc value)]! !

!Prolog methodsFor: 'comm with smalltalk' stamp: 'spfa 6/8/2008 21:30'!
primlogicIs: assoc
        "Execute the second argument as a Smalltalk
         expression and then unify its result with
         the first argument."
    assoc key size = 2 ifFalse: [^self].
    self unify: assoc key head
        with: (self value: assoc key tail head) value
        then: [^assoc value]! !

!Prolog methodsFor: 'comm with smalltalk' stamp: 'spfa 6/8/2008 21:30'!
primlogicRuntime: assoc
        "Unify the first argument with
         the system time in milliseconds."
    assoc key size = 1 ifFalse: [^self].
    self unify: assoc key head
        with: Time millisecondClockValue
        then: [^assoc value]! !


!Prolog methodsFor: 'backtracking' stamp: 'spfa 6/8/2008 21:30'!
primlogicCut: assoc 
	" Succeed always, exit the containing method
	  during backtacking."
	| blk mark |
	assoc key size = 1 ifFalse: [^ self].
	mark _ unboundVars size.
	blk _ assoc value.
	[(blk _ blk value) class == BlockContext] whileTrue.
	self unbindVar: mark.
	assoc key head value! !

!Prolog methodsFor: 'backtracking' stamp: 'spfa 6/8/2008 21:30'!
primlogicExit: assoc
	" exit Prolog "
	| blk |
	blk _ assoc value.
	[(blk _ blk value) class == BlockContext] whileTrue.
	^returnBlock value! !


!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:30'!
primlogicEquiv: assoc
        "Succeed if the two arguments are the same object."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) == (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:30'!
primlogicEq: assoc
        "Succeed if the two arguments can be unified."
    assoc key size = 2 ifFalse: [^self].
    self unify: assoc key head
        with: assoc key tail head
        then: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:31'!
primlogicGe: assoc
        "Succeed if the first argument is
         greater than or equal to the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) >= (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:31'!
primlogicGt: assoc
        "Succeed if the first argument is
         greater than the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) > (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:31'!
primlogicLe: assoc
        "Succeed if the first argument is
         less than or equal to the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) <= (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:31'!
primlogicLt: assoc
        "Succeed if the first argument is
         less than the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) < (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:31'!
primlogicNe: assoc
        "Succeed if the first argument is
         greater than the second."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) ~= (self second: assoc)
        ifTrue: [^assoc value]! !

!Prolog methodsFor: 'comparing' stamp: 'spfa 6/8/2008 21:31'!
primlogicNotEquiv: assoc
        "Succeed if the two arguments
         are not the same object."
    assoc key size = 2 ifFalse: [^self].
    (self first: assoc) == (self second: assoc)
        ifFalse: [^assoc value]! !


!Prolog methodsFor: 'success-failure' stamp: 'spfa 6/8/2008 21:31'!
primlogicFail: assoc
        "Fail always."! !

!Prolog methodsFor: 'success-failure' stamp: 'spfa 6/8/2008 21:31'!
primlogicTrue: assoc
        "Succeed always."
    assoc key size = 0 ifFalse: [^self].
    ^assoc value! !


!Prolog methodsFor: 'debugging' stamp: 'spfa 6/8/2008 21:31'!
primlogicHalt: assoc
        "Halt the execution.  Smalltalk debugger
         can be invoked at this point."
    assoc key size = 0 ifFalse: [^self].
    self halt! !


!Prolog methodsFor: 'input-output' stamp: 'spfa 6/8/2008 21:31'!
primlogicNl: assoc
        "Write an end of line to
         Transcript."
    assoc key size = 0 ifFalse: [^self].
    Transcript cr.
    ^assoc value! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 6/8/2008 21:32'!
primlogicRead: assoc
        "Read a Smalltalk expression, evaluate it and
         answer the result.  This can be used to input
         a Symbol, a number, a quoted string, etc."
    | anAnswer |
    assoc key size = 1 ifFalse: [^self].
    anAnswer := FillInTheBlank
		request: 'type a line and return'
		initialAnswer: ''.
    self unify: assoc key head
        with: anAnswer
        then: [^assoc value]! !

!Prolog methodsFor: 'input-output' stamp: 'spfa 6/8/2008 21:32'!
primlogicWrite: assoc
        "Write all the arguments to Transcript."
    | value |
    assoc key do: [:argument |
        value := self allValue: argument.
        Transcript nextPutAll:
            ((value isKindOf: String)
                ifTrue: [value]
                ifFalse: [value printString])].
    ^assoc value! !


!Prolog methodsFor: 'structures' stamp: 'spfa 6/8/2008 21:28'!
logicArg

arg(n, structure, x) :-
	integer(n),
	is(true, structure value isKindOf: List),
	is(x, (self allValue: structure)
		at: n + 1).! !

!Prolog methodsFor: 'structures' stamp: 'spfa 6/8/2008 21:28'!
logicFunctor

functor(structure, head, arguments) :-
	var(structure),
	atom(head),
	integer(arguments),
	is(structure, self
		structureHead: head
		arguments: arguments).
	"The first argument can be either a
	Relation or a List whose first
	element is a Symbol."
functor([sHead, sTail], head, arguments) :-
	atom(sHead),
	eq(head, sHead),
	is(arguments, sTail value size).! !

!Prolog methodsFor: 'structures' stamp: 'spfa 6/8/2008 21:28'!
logicUniv

univ(structure, list) :-
	nonvar(structure),
	is(true, structure value class == Relation),
	is(x, self allValue: structure),
	is(list, List
		head: x value head
		tail: x value tail).
univ(structure, [head, list]) :-
	atom(head),
	is(x, self allValue: list),
	is(structure, Relation
		head: head value
		tail: x value).! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Prolog class
	instanceVariableNames: ''!

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 7/12/1999 19:22'!
acceptsLoggingOfCompilation
	^true! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 7/7/1999 20:15'!
compilerClass
	^LCompiler! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 8/5/1999 16:21'!
evaluatorClass
	"Answer an evaluator class appropriate for evaluating expressions in the 
	context of this class."

	^LCompiler! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'spfa 6/8/2008 21:20'!
sourceCodeTemplate
	"Answer an expression to be edited and evaluated in order to define 
	methods in this class."

    ^ Logic prologKeyword, 'Predicate
"comment"
predicate(term1, term2) :-
    goal1(term3),
    goal2(term4).'! !

!Prolog class methodsFor: 'prolog-compilation' stamp: 'bolot 7/12/1999 19:22'!
wantsChangeSetLogging
	^true! !


Prolog subclass: #DFAutomata
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog'!

!DFAutomata methodsFor: 'as yet unclassified' stamp: 'spfa 6/8/2008 21:27'!
logicMember

member(x,[x|_]).
member(x,[_|r]) :- member(x,r).! !

!DFAutomata methodsFor: 'as yet unclassified' stamp: 'spfa 6/8/2008 21:27'!
logicParse

parse(L) :- start(S), 
	trans(S,L).! !

!DFAutomata methodsFor: 'as yet unclassified' stamp: 'spfa 6/8/2008 21:27'!
logicTrans

trans(X,[A|B]) :- 
      delta(X,A,Y),   
      write(X),
      write('  '),
      write([A|B]),
      nl(),
      trans(Y,B).  
trans(X,[]) :- 
      final(X),
      write(X),
      write('  '),
      write([]), nl().
! !


Prolog subclass: #Doctor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog'!
!Doctor commentStamp: 'bolot 12/13/2002 15:59' prior: 0!
A demo class.

Used by Family>>health:

This class and other related classes in this category were originally written by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.
We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering.!


!Doctor methodsFor: 'as yet unclassified' stamp: 'spfa 6/8/2008 21:27'!
logicHealthHistory

healthHistory('John', 'John had two times of heart attacks').
healthHistory('Mary', 'Mary had no serious problems').! !


Prolog subclass: #Family
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog'!
!Family commentStamp: 'bolot 12/13/2002 15:24' prior: 0!
A demo class
Examples:
LCompiler evaluate: 'Family new :? father(x,y)'
LCompiler evaluate: 'Family new :? grandPa(''John'',y)'

This class and other related classes in this category were originally written by Mike Teng <miket@objectshare.com>, then at Digitalk, as part of the Prolog/V implementation distributed with Smalltalk V/286 in 1988.
We gratefully acknowledge Mike's permission to port his code to Squeak.

The port to Squeak is the work of Bolot Kerimbaev <bolot@cc.gatech.edu>, a summer intern at Walt Disney Imagineering.!


!Family methodsFor: 'demo' stamp: 'spfa 6/8/2008 21:27'!
logicFather

father('John', 'Mary').  "John is the father of Mary"
father('John', 'David').
father('David', 'Jack').
father('Arthur', 'Nancy').
! !

!Family methodsFor: 'demo' stamp: 'spfa 6/8/2008 23:03'!
logicGrandPa

"x is the grandPa of y."
grandPa(x, y) :-
	father(x, z),
	or(father(z, y),mother(z,y)).! !

!Family methodsFor: 'demo' stamp: 'spfa 6/8/2008 21:28'!
logicHealth

"Return the health history of x."
health(x, history) :-
	consult( healthHistory(x, history), Doctor new).! !

!Family methodsFor: 'demo' stamp: 'spfa 6/8/2008 21:28'!
logicMother

"Mary is the mother of Nancy."
mother('Mary', 'Nancy').! !

!Family methodsFor: 'demo' stamp: 'spfa 6/8/2008 23:03'!
logicParent

parent(x,y) :-
	or(father(x,y), mother(x,y)).! !


Workspace subclass: #PrologWorkspace
	instanceVariableNames: 'doItReceiverPrototype'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prolog'!
!PrologWorkspace commentStamp: 'bolot 12/13/2002 15:47' prior: 0!
PrologWorkspace

enables "pure" Prolog syntax to be used, instead of directly invoking LCompiler

try these:

Family new :? father(x,y)
Family new :? grandPa('John',y)
!


!PrologWorkspace methodsFor: 'as yet unclassified' stamp: 'bolot 12/13/2002 15:46'!
doItReceiver
	^doItReceiverPrototype ifNil: [doItReceiverPrototype _ Prolog new]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PrologWorkspace class
	instanceVariableNames: ''!

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'bolot 12/13/2002 15:58'!
demoFamilyString
	^
'Family new :? father(x,y)
Family new :? grandPa(''John'',y)'! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'spfa 6/9/2008 09:06'!
demoString
	^
'Family new :? father(x,y)
Family new :? grandPa(''John'',y)

Family new :? father(x, ''Mary'')
Family new :? father(''John'', ''Mary'')

Family new :? grandPa(''John'', x), father(y, x)

Family new :? health(''John'', y).
Family new :? consult( healthHistory(''John'', x), Doctor new)


Dungeon new :? go()


Resources:

Prolog Tutorial
http://www.csupomona.edu/~jrfisher/www/prolog_tutorial/contents.html
'! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'bolot 12/13/2002 15:45'!
open
	self openLabel: 'Prolog Workspace'! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'spfa 6/9/2008 09:04'!
openDemo
	"self openDemo"
	(self new contents: self demoString) openLabel: 'Prolog Workspace'! !

!PrologWorkspace class methodsFor: 'as yet unclassified' stamp: 'bolot 12/13/2002 16:09'!
openDemoFamily
	"self openDemoFamily"
	(self new contents: self demoFamilyString) openLabel: 'Prolog Workspace'! !
